#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# cherrytree repairer 
# Version: 1.0.0
# Date:    19.09.2011
# Author:  Bystroushaak (bystrousak@kitakitsune.org)
#
# This work is licensed under a Creative Commons 3.0 
# Unported License (http://creativecommons.org/licenses/by/3.0/cz/).
# Created in Geany text editor.
#
# Notes:
    # 
#===============================================================================
# Imports ======================================================================
#===============================================================================
import os.path
import sys
import argparse

try:
	import dhtmlparser as d
except:
	sys.stderr.write("This module require dhtmlparser.\n")
	sys.stderr.write("You can get it from https://github.com/Bystroushaak/DHTMLParser/tree/python\n")
	sys.exit(1)



#===============================================================================
#= Main program ================================================================
#===============================================================================
parser = argparse.ArgumentParser()
parser.add_argument("filename", help = "File which will be fixed.")
args = parser.parse_args()

# read data
if not os.path.exists(args.filename):
	sys.stderr.write("Can't read '" + str(args.filename) + "'!\n")
	sys.exit(1)
else:
	f = open(args.filename)
	data = f.read()
	f.close()

# parse dom
dom = d.parseString(data.replace("<br>", "\n"))

# remove style from headers
for h in dom.find("h2") + dom.find("h1"):
	if "style" in h.params:
		del h.params["style"]

# replace <span style="font-weight:bolder;"> with <strong>
for span in dom.find("span", {"style":"font-weight:bolder;"}):
	n_span = d.HTMLElement("<strong>")
	n_span.childs = span.childs
	n_span.endtag = d.HTMLElement("</strong>")
	span.replaceWith(n_span)
	
# replace <span style="font-style:italic;"> with <i>
for span in dom.find("span", {"style":"font-style:italic;"}):
	n_span = d.HTMLElement("<i>")
	n_span.childs = span.childs
	n_span.endtag = d.HTMLElement("</i>")
	span.replaceWith(n_span)


def add_p(content):
	c = d.HTMLElement(content)
	
	out = []
	for p in c.__str__().split("\n\n"):
		if p.strip() != "":
			out.append("<p>")
			out.append(p.strip())
			out.append("</p>\n")
	
	return d.parseString("".join(out)).childs

# Use headers as separator. Everything between separator split with "\n\n", add <p> and join.
body = dom.find("body")[0]
alt_body = d.HTMLElement("<body>")
alt_body.endtag = d.HTMLElement("</body>")
content = []
for el in body.childs:
	if el.isTag() and el.getTagName() == "h1" or el.getTagName() == "h2":
		if content != []:
			alt_body.childs.extend(add_p(content))
			content = []
		alt_body.childs.append(el)
	else:
		content.append(el)

if content != []:
	alt_body.childs.extend(add_p(content))

body.replaceWith(alt_body)

print dom.__str__().replace("</p>", "</p>\n")